# Lyapunov_2d_eq_draw.py
#
# Reads the trained neural network "Lyapunov_2d_eq.h5" generated by
# "Lyapunov_2d_eq.py" and plots the result
#
# (C) Lars Grüne 2020

###### import necessary modules ######

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

import tensorflow as tf

from tensorflow import keras

import tensorflow.keras.backend as kb

import matplotlib.pyplot as plt

from mpl_toolkits import mplot3d

# define the vector field
def vf(x):
    y = [-x[:,0]-10.*x[:,1]**2,-2.*x[:,1]]
    return y

# load model
model = keras.models.load_model('Lyapunov_2d_eq.h5')

# define resolution and dimension
numpoints = 30
inputdim = 2

# define plotting range and mesh
x = np.linspace(-1, 1, numpoints)
y = np.linspace(-1, 1, numpoints)

X, Y = np.meshgrid(x, y)

s = X.shape

Ze = np.zeros(s)
Zp = np.zeros(s)
DT = np.zeros((numpoints**2,inputdim))

# convert mesh into point vector for which the model can be evaluated
c = 0
for i in range(s[0]):
    for j in range(s[1]):
        DT[c,0] = X[i,j]
        DT[c,1] = Y[i,j]
        c = c+1;

# evaluate model (= Lyapunov function values V)
Ep = model.predict(DT)

# convert point vector to tensor for evaluating x-derivative
tDT = tf.convert_to_tensor(DT, dtype=tf.float32)

# evaluate gradients DV of Lyapunov function
with tf.GradientTape() as tape:
    tape.watch(tDT)
    ypm = model(tDT)
    grads = tape.gradient(ypm, tDT)

# compute orbital derivative DVf
Ee = kb.sum(grads*kb.transpose(tf.convert_to_tensor(vf(DT), dtype=tf.float32)), axis=1)

# copy V and DVf values into plottable format
c = 0
for i in range(s[0]):
    for j in range(s[1]):
        Ze[i,j] = Ee[c]
        Zp[i,j] = Ep[c]
        c = c+1;

# define figure
fig = plt.figure()
ax = plt.axes(projection='3d')

ax.set_xlabel('$x_1$',fontsize=20)
ax.set_ylabel('$x_2$',fontsize=20)
ax.set_zlabel('$DWf, W$',fontsize=20);

# plot values V
ax.plot_surface(X, Y, Zp, rstride=1, cstride=1,
                cmap='viridis', edgecolor='none')

# plot orbital derivative DVf
ax.plot_wireframe(X, Y, Ze, rstride=1, cstride=1)

plt.show()
